# python 3, export all objects to an arbitrary xml file
# format. This only works with Blender 2.6+

bl_info = {
    "name": "Export Generic Game XML",
    "author": "Execution Unit Ltd",
    "version": (1, 0),
    "blender": (2, 5, 7),
    "location": "File > Export > GameXML (.xml)",
    "description": "Export game xml file (.xml)",
    "warning": "",
    "wiki_url": "",
    "tracker_url": "",
    "category": "Import-Export"}

import bpy
import bmesh
import re
from datetime import datetime
from collections import OrderedDict

instre = re.compile(r"(?P<name>[^.]*)\.\d+\Z")
materials = OrderedDict()


class Material(object):
    """An instance of a material. It copies out useful information and
    should be put in to the materials OrderedDict."""
    def __init__(self, bmat):
        """Create a material from a blender material object"""
        super().__init__()
        self.name = bmat.name
        self.textures = []
        for tslot in bmat.texture_slots:
            if tslot is None:
                continue
            tex = tslot.texture
            if tex and tex.type == "IMAGE":
                self.textures.append((tex.name, tex.image.filepath, tex.extension))

    def export(self, f):
        f.write("<material type='dict'>")
        f.write("<name type='string'>%s</name>" % self.name)
        f.write("<textures type='list'>")
        for idx, tex in enumerate(self.textures):
            f.write("<texture type='dict'>")
            f.write("<relpath type='string'>%s</relpath>" % (tex[1]))
            f.write("<extension type='string'>%s</extension>" % tex[2])
            f.write("</texture>")
        f.write("</textures>")
        f.write("</material>")


def _isinstance_getname(obj):
    """return true, name if an object is an instance of another"""
    name = obj.name
    instance = False
    match = instre.search(obj.name)
    if match:
        name = match.group("name")
        instance = True
    return instance, name


def _isinstance(obj):
    """return True if the object obj is an instance of another"""
    i, _ = _isinstance_getname(obj)
    return i


def _gathermaterials(context):
    """iterate through all objects gathering together information about
    materials so we can export a single set of materials for all models"""
    objects = context.scene.objects
    for obj in objects:
        if obj.type != "MESH" or _isinstance(obj):
            continue
        mesh = obj.data
        for mat in mesh.materials:
            if mat.name not in materials:
                materials[mat.name] = Material(mat)


def _exportmeshinstance(f, name, obj):
    f.write("<%s type='dict'>" % obj.name)
    f.write("<class type='string'>instance</class>")
    f.write("<instance type='bool'>%d</instance>" % 1)
    f.write("<source type='string'>%s</source>" % name)
    _exportobjectdata(f, obj)
    f.write("</%s>" % obj.name)


def _getpropertytype(propvalue):
    ops = {"int": int, "float": float, "string": str}
    for name, op in ops.items():
        if type(propvalue) == op:
            return name
        # try:
        #     dummy = op(propvalue)
        #     return name
        # except ValueError:
        #     pass
    return "string"


def _exportobjectdata(f, obj):
    l = obj.location
    f.write("<pos type='vector3'>%f,%f,%f</pos>" % (l[0], l[1], l[2]))

    # make sure we get the rotation as EulerXYZ. this isn't the best way to
    # do it in general but for my purposes it's fine.
    oldrotationmodel = obj.rotation_mode
    obj.rotation_mode = "XYZ"
    rot = obj.rotation_euler
    obj.rotation_mode = oldrotationmodel
    # make sure we write in 'correct order' x,y,z,w
    f.write("<rot type='vector3'>%f,%f,%f</rot>" % (rot[0], rot[1], rot[2]))

    s = obj.scale
    f.write("<scale type='vector3'>%f,%f,%f</scale>" % (s[0], s[1], s[2]))

    f.write("<localbounds type='list'>")
    for name, i in [("min", 0), ("max", 6)]:
        f.write("<%s type='vector3'>%f,%f,%f</%s>" %
            (name, obj.bound_box[i][0], obj.bound_box[i][1], obj.bound_box[i][2], name))
    f.write("</localbounds>")

    f.write("<properties type='dict'>")
    for item in obj.items():
        if not item[0].startswith('_'):
            proptype = _getpropertytype(item[1])
            f.write("<%s type='%s'>%s</%s>" % (item[0], proptype, item[1], item[0]))
    f.write("</properties>")


def _exportmeshmaterials(f, obj, mesh):
    """write the materials in the mesh object"""
    if len(mesh.materials) == 0:
        raise Exception("Mesh %s has no material set\n" % obj.name)
    f.write("<materials type='list'>")
    for idx, mat in enumerate(mesh.materials):
        f.write("<material type='dict'>")
        f.write("<id type='int'>%d</id>" % idx)
        f.write("<name type='string'>%s</name>" % mat.name)
        f.write("</material>")
    f.write("</materials>")


def _exportmeshvertices(f, mesh):
    """write the vertices of mesh mesh in to file f"""
    verts = mesh.vertices
    f.write("<vertices type='list'>")
    for vert in verts:
        f.write("<vert type='dict'>")
        f.write("<pos type='vector3'>%f,%f,%f</pos>" %
               (vert.co[0], vert.co[1], vert.co[2]))
        if(len(vert.groups)):
            f.write("<vertexgroups type='list'>")
            for grp in vert.groups:
                f.write("<vertexgroup type='dict'>")
                f.write("<index type='int'>%d</index>" % grp.group)
                f.write("<weight type='float'>%f</weight>" % grp.weight)
                f.write("</vertexgroup>")
            f.write("</vertexgroups>")
        f.write("</vert>")
    f.write("</vertices>")


def _exportmeshfaces(f, mesh):
    f.write("<faces type='list'>")

    uv_layer = mesh.uv_layers.active.data if mesh.uv_layers.active else None
    col_layer = mesh.vertex_colors.active.data if mesh.vertex_colors.active else None

    for poly in mesh.polygons:
        f.write("<face type='dict'>")
        # range is used here to show how the polygons reference loops,
        # for convenience 'poly.loop_indices' can be used instead.
        f.write("<vertices type='list'>")
        for loop_index in range(poly.loop_start, poly.loop_start + poly.loop_total):
            f.write("<vert type='dict'>")
            f.write("<index type='int'>%d</index>" % (mesh.loops[loop_index].vertex_index))
            if uv_layer:
                uv = uv_layer[loop_index].uv
                f.write("<uv type='vector2'>%f,%f</uv>" % (uv[0], uv[1]))
            if col_layer:
                c = col_layer[loop_index].color
                f.write("<col type='vector3'>%f,%f,%f</col>" % (c[0], c[1], c[2]))
            f.write("</vert>")
        f.write("</vertices>")
        f.write("<material type='int'>%d</material>" % poly.material_index)
        f.write("<normal type='vector3'>%f, %f, %f</normal>" % \
                (poly.normal[0], poly.normal[1], poly.normal[2]))
        f.write("</face>")

    f.write("</faces>")


def _exportvertexgroups(f, obj):
    f.write("<vertexgroups type='list'>")
    for vg in obj.vertex_groups:
        f.write("<vertexgroup type='dict'>")
        f.write("<index type='int'>%d</index>" % vg.index)
        f.write("<name type='string'>%s</name>" % vg.name)
        f.write("</vertexgroup>")
    f.write("</vertexgroups>")


def _export_mesh(f, obj):
    """export object obj to file f"""
    mesh = obj.data
    f.write("<%s type='dict'>" % obj.name)
    f.write("<class type='string'>mesh</class>")
    _exportobjectdata(f, obj)
    _exportmeshmaterials(f, obj, mesh)
    _exportmeshvertices(f, mesh)
    _exportvertexgroups(f, obj)
    _exportmeshfaces(f, mesh)
    f.write("</%s>" % obj.name)


def _exportempty(f, name, obj):
    f.write("<%s type='dict'>" % obj.name)
    f.write("<class type='string'>empty</class>")
    _exportobjectdata(f, obj)
    f.write("</%s>" % obj.name)


def _exportscene(f, scene):
    f.write("<scene type='dict'>")
    objects = scene.objects
    for obj in objects:
        if not obj.is_visible(scene):
            continue
        name = obj.name
        instance = False
        match = instre.search(obj.name)
        if match:
            name = match.group("name")
            instance = True

        t = obj.type
        if t == 'MESH':
            if instance:
                _exportmeshinstance(f, name, obj)
            else:
                _export_mesh(f, obj)
        elif t == "EMPTY":
            _exportempty(f, name, obj)

    f.write("</scene>")


def _exportmaterials(f):
    """Export the materials gathered in _gathermaterials stored in the
    materials OrderedDict"""
    f.write("<materials type='list'>")
    for name, mat in materials.items():
        mat.export(f)
    f.write("</materials>")


def _exportheader(f, context):
    """export general information about the blendfile"""
    f.write("<header type='dict'>")
    f.write("<blendfile type='path'>%s</blendfile>" % bpy.data.filepath)
    f.write("<date type='date'>%s</date>" % str(datetime.now()))
    f.write("</header>")


def _doexport(context, filepath):
    f = open(filepath, 'w', encoding='utf-8')
    f.write("<?xml version=\"1.0\"?>")
    f.write("<blenderexport type='dict'>")
    _gathermaterials(context)
    _exportheader(f, context)
    _exportmaterials(f)
    _exportscene(f, context.scene)
    f.write("</blenderexport>")
    f.close()
    return {'FINISHED'}


# ExportHelper is a helper class, defines filename and
# invoke() function which calls the file selector.
from bpy_extras.io_utils import ExportHelper
from bpy.props import StringProperty, BoolProperty, EnumProperty
from bpy.types import Operator


class ExUnitXMLExport(Operator, ExportHelper):
    '''This appears in the tooltip of the operator and in the generated docs'''
    # important since its how bpy.ops.import_test.some_data is constructed
    bl_idname = "exunit.exunitxmlexport"
    bl_label = "ExUnit XML"

    # ExportHelper mixin class uses this
    filename_ext = ".xml"

    filter_glob = StringProperty(
            default="*.xml",
            options={'HIDDEN'},
            )

    # List of operator properties, the attributes will be assigned
    # to the class instance from the operator settings before calling.
    # use_setting = BoolProperty(
    #         name="Example Boolean",
    #         description="Example Tooltip",
    #         default=True,
    #         )

    # type = EnumProperty(
    #         name="Example Enum",
    #         description="Choose between two items",
    #         items=(('OPT_A', "First Option", "Description one"),
    #                ('OPT_B', "Second Option", "Description two")),
    #         default='OPT_A',
    #         )

    def execute(self, context):
        return _doexport(context, self.filepath)


# Only needed if you want to add into a dynamic menu
def menu_func_export(self, context):
    self.layout.operator(ExUnitXMLExport.bl_idname, text="ExUnit Export")


def register():
    bpy.utils.register_class(ExUnitXMLExport)
    bpy.types.INFO_MT_file_export.append(menu_func_export)


def unregister():
    bpy.utils.unregister_class(ExUnitXMLExport)
    bpy.types.INFO_MT_file_export.remove(menu_func_export)


if __name__ == "__main__":
    register()

    bpy.ops.exunit.exunitxmlexport('INVOKE_DEFAULT')
